----------------------------------
-- Create plugin
local RW = LibStub("AceAddon-3.0"):GetAddon("Raid Watch")
local plugin = RW:Plugin("Controller")
if not plugin then return end

----------------------------------
-- Locals 

local L = LibStub("AceLocale-3.0"):GetLocale("RW2-Plugin-Controller")
local LB = LibStub("AceLocale-3.0"):GetLocale("RW2-Plugin-Base")
local LBB = LibStub("AceLocale-3.0"):GetLocale("RW2-Base")
local LSM = LibStub("LibSharedMedia-3.0")
local AGU = LibStub("AceGUI-3.0")
local FA = LibStub("LibFrameAnchorRegistry-1.0")
local UIF

local db
local Utils = RW.Utils

local Window, ConfigWindow, BossConfigButton, Anchor
local StatusString, TimeString
local CurrentBoss, ContentHeight, ConfigMode
local dragStart, dragStop, BossFrameUpdate
local Party
local pGUID = UnitGUID("player")

local defaults = {
	profile = {
		PluginEnabled = true,
		AnchorFrame = nil,
		Point = "CENTER",
		RelPoint = "CENTER",
		x = 0, 
		y = 0,
		Width = 300,
		
		ShowOnEngage = true,
		ShowOnLoad = true,
		ShowOnZone = true,
		ShowOnJoin = true,
				
		FontSettings = {
			Font = "accid",
			Size = 16,
			Flags = "NONE",
			Color = {1, 1, 1, 1},
		},
		
		BackgroundColor = {0, 0, 0, 0.5},
		Border = "None",
		BorderWidth = 10,
		BorderColor =  {1, 1, 1, 1},
		Inset = 3,
	}
}

----------------------------------
-- Init

function plugin:OnRegister()
	self.db = RW.db:RegisterNamespace(self:GetName(), defaults)
	db = self.db.profile
	
	-- DB Upgrade
	for name, profile in pairs(self.db.profiles) do
		local db = profile
		db.FontSettings = db.FontSettings or {}
		if db.Font then	db.FontSettings.Font = db.Font end
		if db.FontSize then db.FontSettings.Size = db.FontSize end
		if db.FontFlags then db.FontSettings.Flags = db.FontFlags end
		
		db.Font = nil
		db.FontSize = nil
		db.FontFlags = nil
		db.FontColor = nil	
	end	
	
	self.db.RegisterCallback(self, "OnProfileChanged", "Reset")
	self.db.RegisterCallback(self, "OnProfileCopied", "Reset")
	self.db.RegisterCallback(self, "OnProfileReset", "Reset")
end

function plugin:PluginEnable()
	if not db.PluginEnabled then
		self:Disable()
		return
	end
	
	UIF = LibStub("LibGUIFactory-1.0"):GetFactory("RW")
	UIF:SetValuesChangedCallback(function()
		if ConfigWindow and ConfigWindow.frame:IsShown() then
			ConfigWindow:ReleaseChildren()
			ConfigWindow:AddChild(plugin:GetGUI())
		end
	end)
	
	FA.RegisterCallback(self, "FrameRegistered", "ApplySettings")
	
	Party = RW.Party
	
	self:RegisterCallback("BossEngaged")
	self:RegisterCallback("BossDisengaged")

	self:RegisterCallback("ConfigMode")
	self:RegisterCallback("EnteringEnabledZone")
	self:RegisterCallback("LeavingEnabledZone")
	self:RegisterCallback("ToggleController")
	
	self:CreateWindow()
	if db.ShowOnLoad then		
		Utils.FadeIn(Window)
	end	
end

function plugin:PluginDisable()
	FA.UnregisterAllCallbacks(self)
end

function plugin:Reset()
	db = self.db.profile
	self:ApplySettings()
end

function UpdateTitle(...)
	local str = ""
	if CurrentBoss then
		str = RW.LBB[CurrentBoss:GetName()] or CurrentBoss:GetName()
	else
		str = date("%H:%M:%S")
	end
	local partyType = Party:GetPartyType()
	if partyType == "raid" or partyType == "party" then
		str = ("%s - %d %s"):format(str, Utils:InstanceSize(), (Utils:IsHeroic() and "Heroic" or "Normal"))
	end
	Window:SetTitle(str)
end

function plugin:CreateWindow()
	local f = Utils:CreateAnchor(name, UIParent, "Controller")
	f.name = name
	f:SetScript("OnDragStart", dragStart)
	f:SetScript("OnDragStop", dragStop)
	f:SetPoint("CENTER", db.x, db.y)
	f:SetWidth(db.Width)	
	Anchor = f
	
	if not Window then	
		Window = Utils:CreateWindow("Controller", true, dragStart, dragStop)
		Window:HideContent()
		Window:SetPoint("BOTTOMLEFT", f, "TOPLEFT")
		Window:SetPoint("BOTTOMRIGHT", f, "TOPRIGHT")
		Window:ShowLeftIcon([[Interface\Icons\Ability_Mage_HotStreak]])
		Window:SetAlpha(0)
		Window:SetTitle("")
		Utils:RegisterOnUpdate(Window, UpdateTitle, 1)
		
		local b = Window:AddTitleButton("Close", nil, true)	
		b:AddClickHandler(function() 
			Utils.FadeOut(Window)
		end)
		Utils:AddTooltip(b, LB.CLOSE)
		
		b = Window:AddPulloutButton("RangeCheck")	
		b:AddClickHandler(function() 
			RW.Callbacks:Fire("RangeCheckShow", true, nil, true) 
		end)
		Utils:AddTooltip(b, "Show Range Check")
		
		b = Window:AddPulloutButton("DKP")	
		b:AddClickHandler(function() 
			RW.Callbacks:Fire("DKPShow")			
		end)
		Utils:AddTooltip(b, "Show DKP")
		
		b = Window:AddPulloutButton("Versioncheck")
		b:AddClickHandler(function()
			RW.Callbacks:Fire("VersionsShow")			
		end)
		Utils:AddTooltip(b, "Show Versions")
		
		b = Window:AddPulloutButton("Config")
		b:AddClickHandler( function()
			plugin:OpenConfigWindow()
		end)
		Utils:AddTooltip(b, LB.CONFIGURE)
		
		Window:Collapse()
		
		StatusString = Window.content:CreateFontString(nil, "ARTOWRK", "GameFontNormal")
		StatusString:SetPoint("BOTTOMLEFT", 2, 2)
		TimeString = Window.content:CreateFontString(nil, "ARTOWRK", "GameFontNormal")
		TimeString:SetPoint("TOPLEFT", 2, -2)
	end	
	Window:Hide()
	self:ApplySettings()
	FA:RegisterAnchor("Raid Watch", "Controler Title", "RaidWatchControlerTitle", Window)
end

function plugin:CreateBossWindow()
	BossWindow = CreateFrame("Frame", nil, Window.content)
	BossWindow:SetPoint("BOTTOMLEFT", Window.content)
	BossWindow:SetPoint("BOTTOMRIGHT", Window.content)
	BossWindow:EnableMouse(true)
	BossWindow:SetScript("OnEnter", function() Utils.FadeIn(BossConfigButton) end)
	BossWindow:SetScript("OnLeave", function() Utils.FadeOut(BossConfigButton) end)
	
	local t = BossWindow:CreateFontString(nil, nil, "GameFontNormal")
	t:SetJustifyH("LEFT")
	t:SetPoint("TOPLEFT", 5, -5)
	BossWindow.timestr = t
	
	t = BossWindow:CreateFontString(nil, nil, "GameFontNormal")
	t:SetJustifyH("LEFT")
	t:SetPoint("TOPLEFT", BossWindow.timestr, "BOTTOMLEFT")
	BossWindow.partystatus = t
	
	local b = Window:AddTitleButton("Config")
	b:AddClickHandler(function(self, button, isDown)
		if not IsAddOnLoaded("RaidWatch_Options") and not RW:GetPlugin("Options") then
			if not RWL:LoadAddOn("RaidWatch_Options") then
				return
			end
		end		
		local options = RW:GetPlugin("Options")
		if options then
			local frame = UIF:Frame(RW.LBB[BossWindow.currentBoss:GetName()] or BossWindow.currentBoss:GetName())
			frame:SetLayout("Fill")
			frame:Show()
			
			local scroll = AGU:Create("ScrollFrame")
			frame:AddChild(scroll)
			scroll:AddChild(options.SelectBoss(BossWindow.currentBoss, frame))						
			
			scroll:DoLayout()
			scroll:FixScroll()
			frame:DoLayout()
		else
			self:Meg("Options is enabled, options for "..(RW.LBB[BossWindow.currentBoss:GetName()] or BossWindow.currentBoss:GetName()).." cannot be shown")
		end
	end)
	b:SetPoint("RIGHT", BossWindow, -5, 0)
	b:SetParent(BossWindow)
	b:Hide()
	BossConfigButton = b
end

local ConfigStatus = {}
function plugin:OpenConfigWindow(toggle)
	if not ConfigWindow then
		ConfigWindow = UIF:Frame(L.NAME)
		ConfigWindow:SetCallback("OnClose", function()
			ConfigWindow:ReleaseChildren()
			ConfigWindow.frame:Hide()
			ConfigWindow = nil
		end)
		ConfigWindow:SetLayout("Fill")	
		ConfigWindow:Show()
		ConfigWindow:SetStatusTable(ConfigStatus)
		
		local scroll = AGU:Create("ScrollFrame")
		scroll:SetLayout("Flow")
		ConfigWindow:AddChild(scroll)
		
		scroll:AddChildren(plugin:GetGUI())
		Utils:DoLayout(ConfigWindow)
		scroll:FixScroll()
		return
	else
		ConfigWindow:Release()
		ConfigWindow = nil
	end
end

function plugin:ApplySettings()
	self = plugin
	
	if Window then
		Anchor:SetWidth(db.Width)	
		
		local anchor = FA:GetAnchor(db.AnchorFrame or "BlizzUIParent")
		if not anchor then anchor = UIParent end
		
		Anchor:ClearAllPoints()
		Anchor:SetPoint(db.Point, anchor, db.RelPoint, db.x, db.y)
		Anchor:SetWidth(db.Width)
		
		Window:SetBorder(LSM:Fetch("border", db.Border), db.BorderWidth, db.BorderColor)
		Window.content.background:SetTexture(unpack(db.BackgroundColor))
		Window:SetInsets(db.Border == "None" and 0 or db.Inset)		
	end
	
	TimeString:SetFont(LSM:Fetch("font", db.FontSettings.Font), db.FontSettings.Size, db.FontSettings.Flags)
	TimeString:SetTextColor(unpack(db.FontSettings.Color))
	
	StatusString:SetFont(LSM:Fetch("font", db.FontSettings.Font), db.FontSettings.Size, db.FontSettings.Flags)
	StatusString:SetTextColor(unpack(db.FontSettings.Color))
	
	local oldText = TimeString:GetText()
	TimeString:SetText("IsdgsdgsdgSFSAGASGdg")
	ContentHeight = TimeString:GetStringHeight() * 2 + 10 + (db.Border == "None" and 0 or db.Inset * 2)
	TimeString:SetText(oldText)
	
	if ConfigMode or CurrentBoss then
		Window:SetContentHeight(ContentHeight)
	end
end

-- API ---------------------------------------
----------------------------------------------
function plugin:ToggleController()
	if not Window then self:CreateWindow() end
	
	if Window:IsShown() then 
		Utils.FadeOut(Window)
		Window:Collapse()
	else 
		Utils.FadeIn(Window, true) 
		if ConfigMode or CurrentBoss then
			Window:SetContentHeight(ContentHeight)
		else
			Window:Restore()
		end
	end
end

function plugin:BossEngaged(event, boss)
	CurrentBoss = boss
	if db.ShowOnEngage then
		if not Window then self:CreateWindow() end
		Window:Show()
	end
	
	if Window then		
		Utils:RegisterOnUpdate(Window, BossFrameUpdate, 0.5)
		--Utils:AddTooltip(BossConfigButton, RW.LBB[BossWindow.currentBoss:GetName()] or BossWindow.currentBoss:GetName(),  "Open Config")
		Window:ShowLeftIcon(boss.BossIcon)
		Window:SetContentHeight(ContentHeight)
		Window.start = 0
	end
end

function plugin:BossDisengaged(event, boss, wipe)
	if Window then	
		Window:ShowLeftIcon([[Interface\Icons\Ability_Mage_HotStreak]])
		Window:HideContent()
		Utils:UnregisterOnUpdate(Window, BossFrameUpdate)
		
		if not db.ShowOnZone and not db.ShowOnLoad then
			Utils.FadeOut(Window)
		end
	end
	CurrentBoss = nil
end

function plugin:EnteringEnabledZone(event)
	if not db.ShowOnZone then return end
	if not Window then self:CreateWindow() end	
	Window:Show()
end

function plugin:LeavingEnabledZone(event)
	if db.ShowOnLoad then return end
	if Window then Utils.FadeOut(Window) end
end

function plugin:ConfigMode(event, show)
	ConfigMode = show
	if not Window then self:CreateWindow() end	
	if show then
		Anchor:Show()
		Utils.FadeIn(Window)
		Window:SetContentHeight(ContentHeight)
		TimeString:SetText("Elapsed time here")
		StatusString:SetText("Status here")
		Window:SetContentHeight(ContentHeight)		
	else
		Anchor:Hide()
		Window:HideContent()
	end
end

-- UI ----------------------------------------
----------------------------------------------
do
	function plugin:GetOptions()
		return {{text = LB.PLUGINS, value = "Plugins"}, {text = L.NAME, value = "Controller"}}, self.GetGUI, [[Interface\Icons\Ability_Mage_HotStreak]]
	end
	
	local TabSelected
	local LastSelectedTab
	
	function plugin:GetGUI()
		self = plugin
		
		local scroll = AGU:Create("SimpleGroup")	
		scroll:SetFullWidth(true)
		scroll:SetLayout("Flow")
		
		local label = UIF:MainTitleLabel(L.NAME)
		scroll:AddChild(label)	
		
		-- Information
		local infogroup = UIF:InlineGroup1()
		local label = UIF:Text1(L.DESC)
		label:SetImage([[Interface\Icons\Ability_Mage_HotStreak]])
		infogroup:AddChild(label)
		scroll:AddChild(infogroup)
		
		local group = UIF:InlineGroup1("Visible")
		scroll:AddChild(group)
		
		local checkbox
		group:AddChild(UIF:CheckBox(L.ALWAYS_SHOW, db, "ShowOnLoad", plugin.ApplySettings, L.ALWAYS_SHOW_DESC, 0.5, true))
		group:AddChild(UIF:CheckBox(L.SHOW_BOSSFIGHT, db, "ShowOnEngage", nil, L.SHOW_BOSSFIGHT_DESC, 0.5, true))
		group:AddChild(UIF:CheckBox(L.SHOW_ZONE, db, "ShowOnZone", nil, L.SHOW_ZONE_DESC, 0.5, true))
		group:AddChild(UIF:CheckBox(L.SHOW_JOIN_PARTY, db, "ShowOnJoin", nil, L.SHOW_JOIN_PARTY_DESC, 0.5, true))
		
		local tgroup = UIF:TabGroup1()
		scroll:AddChild(tgroup)
		tgroup:SetTabs(
		{
			{text = LB.POSITION .." & " .. LB.SIZE, value = "Position"},
			{text = LB.TEXTURE_MAIN, value = "Textures"},
			{text = LB.FONT, value = "Font"},
		})
		tgroup:SetCallback("OnGroupSelected", TabSelected)
		tgroup:SelectTab(LastSelectedTab or "Position")					
		
		button = UIF:Button(LB.RESET, function() 
			plugin.db:ResetProfile() 
		end)
		button:SetUserData("confirm", true)
		button:SetUserData("TriggerGlobalUpdate", true)
		scroll:AddChild(button)
		
		return scroll
	end
	
	TabSelected = function(self, event, tab)
		self:ReleaseChildren()
		LastSelectedTab = tab
		
		if tab == "Position" then
			if not Anchor then plugin:CreateWindow() end
			local t = AGU:Create("FrameAnchorDropdown")
			t:SetRelativeWidth(0.5)			
			t:SetList(FA:GetAnchors(Anchor))
			t:SetLabel("Anchor")
			t:SetValue(db.AnchorFrame)
			t:SetCallback("OnValueChanged", function(self, event, cat, name, frame)
				db.AnchorFrame = cat
				plugin.ApplySettings()
			end)	
			self:AddChild(t)
			self:AddChild(UIF:NewLine())
			
			self:AddChild(UIF:GridSelect("Point", db, "Point", "Anchor", 3, plugin.ApplySettings, 0.5, 75))
			self:AddChild(UIF:GridSelect("Relative Point", db, "RelPoint", "Anchor", 3, plugin.ApplySettings, 0.5, 75))
			
			slider = UIF:Slider("X", db, "x", 
				-floor(GetScreenWidth() * UIParent:GetEffectiveScale()), 
				floor(GetScreenWidth() * UIParent:GetEffectiveScale()),
				1, plugin.ApplySettings)
			self:AddChild(slider)
			
			slider = UIF:Slider("Y", db, "y", 
				-floor(GetScreenHeight() * UIParent:GetEffectiveScale()), 
				floor(GetScreenHeight() * UIParent:GetEffectiveScale()),
				1, plugin.ApplySettings)
			self:AddChild(slider)
			self:AddChild(UIF:Slider("Width", db, "Width", 10, 500, 1, plugin.ApplySettings))
			
		elseif tab == "Textures" then
			self:AddChild(UIF:ColorSelect(LB.BACKGROUND_COLOR, db, "BackgroundColor", plugin.ApplySettings, true, 0.5))
			self:AddChild(UIF:ColorSelect(LB.BORDER_COLOR, db, "BorderColor", plugin.ApplySettings, true, 0.5))			
			self:AddChild(UIF:LSMDropdown("border", LB.BORDER_TEXTURE, db, "Border", plugin.ApplySettings, 0.5))			
			self:AddChild(UIF:Slider(LB.BORDER_SIZE, db, "BorderWidth", 1, 40, 1, plugin.ApplySettings, 0.5))			
			self:AddChild(UIF:Slider(LB.INSET, db, "Inset", 0, 50, 1, plugin.ApplySettings, 0.5))
			
		elseif tab == "Font" then
			self:AddChild(UIF:FontSettings(LB.FONT_MAIN, db.FontSettings, true, plugin.ApplySettings, 1))
		end
		
		Utils:DoLayout(self)
	end
end
-- Moving ------------------------------------
----------------------------------------------
dragStart = function(self)
	Anchor:StartMoving()
end

dragStop = function(self)
	Anchor:StopMovingOrSizing()
	local scale = Anchor:GetEffectiveScale() / UIParent:GetEffectiveScale()
	
	local x, y = Anchor:GetCenter()
	x, y = x * scale, y * scale
	
	x = x - GetScreenWidth()/2
	y = y - GetScreenHeight()/2
	x = x * scale
	y = y * scale	
	
	db.x, db.y = x, y
	db.AnchorFrame = nil
	db.Point = "CENTER"
	db.RelPoint = "CENTER"
	
	RW.Callbacks:Fire("RefreshOptions")
	if ConfigWindow then
		plugin:OpenConfigWindow(true)
	end
end

BossFrameUpdate = function(self, elapsed)
	TimeString:SetFormattedText(("%s: %s"):format(LB.DURATION, Utils:TimeToStringShort(self.start)))
	StatusString:SetText(("Status: %d / %d alive"):format(Party:GetNumAlive(), Party:GetPartySize()))
	self.start = self.start + elapsed
end
